Package: MetricGraph
Type: Package
Title: Random Fields on Metric Graphs
Version: 1.6.0
Authors@R: c(
    person("David", "Bolin", email = "davidbolin@gmail.com", role = c("cre", "aut")),
    person("Alexandre", "Simas", email = "alexandre.impa@gmail.com", role = "aut"),
    person("Jonas", "Wallin", email = "jonas.wallin81@gmail.com", role = "aut"))
Maintainer: David Bolin <davidbolin@gmail.com>
Description: Facilitates creation and manipulation of metric graphs, such as street or river networks. Further facilitates operations and visualizations of data on metric graphs, and the creation of a large class of random fields and stochastic partial differential equations on such spaces. These random fields can be used for simulation, prediction and inference. In particular, linear mixed effects models including random field components can be fitted to data based on computationally efficient sparse matrix representations. Interfaces to the R packages 'INLA' and 'inlabru' are also provided, which facilitate working with Bayesian statistical models on metric graphs. The main references for the methods are Bolin, Simas and Wallin (2024) <doi:10.3150/23-BEJ1647>, Bolin, Kovacs, Kumar and Simas (2023) <doi:10.1090/mcom/3929> and Bolin, Simas and Wallin (2023) <doi:10.48550/arXiv.2304.03190> and <doi:10.48550/arXiv.2304.10372>.
License: GPL (>= 2)
Depends: R (>= 4.1.0)
Imports: stats, RANN, ggplot2 (>= 3.4.0), igraph (>= 1.3.0), sf (>=
        1.0-0), rSPDE (>= 2.5.0), Matrix (>= 1.5-0), methods, Rcpp (>=
        1.0.5), R6, lifecycle (>= 1.0.0), sp, dplyr (>= 1.0.0), tidyr
        (>= 1.1.0), magrittr, broom, zoo, ggnewscale, rlang (>= 0.4.0),
        foreach, doParallel, spatstat.geom
Suggests: knitr, testthat, INLA (>= 24.12.01), inlabru (>= 2.14.0),
        osmdata, sn, plotly, parallel, optimParallel, numDeriv, SSN2,
        cowplot, leaflet, mapview, viridis, fmesher (>= 0.6.0),
        data.table, spatstat.data
Additional_repositories: https://inla.r-inla-download.org/R/testing
BugReports: https://github.com/davidbolin/MetricGraph/issues
URL: https://davidbolin.github.io/MetricGraph/
Copyright: The R package and code, and the main programs, were written
        by and are Copyright by David Bolin, Alexandre B. Simas and
        Jonas Wallin, and are redistributable under the GNU Public
        License, version 2 or later. The package also includes partial
        codes from another package, which was deprecated in Oct-2023,
        and whose codes are under the GPL-2 license. For details see
        the COPYRIGHTS file.
Encoding: UTF-8
LazyData: true
VignetteBuilder: knitr
NeedsCompilation: yes
LinkingTo: Rcpp, RcppEigen
Config/roxygen2/version: 8.0.0
Packaged: 2026-05-06 10:39:44 UTC; debusta
Author: David Bolin [cre, aut],
  Alexandre Simas [aut],
  Jonas Wallin [aut]
Repository: CRAN
Date/Publication: 2026-05-06 11:50:12 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-06 12:54:29 UTC; unix
Archs: MetricGraph.so.dSYM
