# Generated by roxygen2: do not edit by hand

S3method(augment,graph_lme)
S3method(deviance,graph_lme)
S3method(drop_na,metric_graph_data)
S3method(filter,metric_graph_data)
S3method(gg_df,metric_graph_spde_result)
S3method(glance,graph_lme)
S3method(inlabru::bru_get_mapper,inla_metric_graph_spde)
S3method(inlabru::ibm_jacobian,bru_mapper_inla_metric_graph_spde)
S3method(inlabru::ibm_n,bru_mapper_inla_metric_graph_spde)
S3method(inlabru::ibm_values,bru_mapper_inla_metric_graph_spde)
S3method(logLik,graph_lme)
S3method(mutate,metric_graph_data)
S3method(nobs,graph_lme)
S3method(plot,graph_bru_pred)
S3method(plot,graph_bru_proc_pred)
S3method(predict,graph_lme)
S3method(predict,inla_metric_graph_spde)
S3method(predict,rspde_metric_graph)
S3method(print,graph_lme)
S3method(print,metric_graph_edge)
S3method(print,metric_graph_edges)
S3method(print,metric_graph_vertex)
S3method(print,metric_graph_vertices)
S3method(print,summary_graph_lme)
S3method(profile,graph_lme)
S3method(select,metric_graph_data)
S3method(simulate,graph_lme)
S3method(summarise,metric_graph_data)
S3method(summary,graph_lme)
S3method(summary,metric_graph)
S3method(summary,metric_graph_spde_result)
export("%>%")
export(Cholesky)
export(augment)
export(cross_validation)
export(drop_na)
export(exp_covariance)
export(filter)
export(gg_df)
export(glance)
export(graph_bru_process_data)
export(graph_components)
export(graph_data_spde)
export(graph_lgcp_sim)
export(graph_lme)
export(graph_spde)
export(graph_spde_basis)
export(graph_spde_make_A)
export(graph_starting_values)
export(lgcp_graph)
export(linnet.to.graph)
export(logo_lines)
export(make_Q_euler)
export(make_Q_spacetime)
export(match_mesh_data)
export(metric_graph)
export(mutate)
export(posterior_crossvalidation)
export(posterior_crossvalidation_loo)
export(precompute_lgcp_graph)
export(process_rspde_predictions)
export(psp.to.graph)
export(sample_spde)
export(select)
export(selected_inv)
export(simulate_spacetime)
export(spde_covariance)
export(spde_metric_graph_result)
export(spde_precision)
export(spde_variance)
export(stlpp.to.graph)
export(summarise)
export(t)
export(update_graph)
import(Matrix)
importFrom(R6,R6Class)
importFrom(RANN,nn2)
importFrom(Rcpp,evalCpp)
importFrom(broom,augment)
importFrom(broom,glance)
importFrom(doParallel,registerDoParallel)
importFrom(dplyr,arrange)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,if_else)
importFrom(dplyr,mutate)
importFrom(dplyr,select)
importFrom(dplyr,summarise)
importFrom(dplyr,ungroup)
importFrom(foreach,`%dopar%`)
importFrom(ggnewscale,new_scale_color)
importFrom(ggplot2,aes)
importFrom(ggplot2,coord_fixed)
importFrom(ggplot2,geom_path)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,guide_legend)
importFrom(ggplot2,labs)
importFrom(ggplot2,scale_colour_discrete)
importFrom(ggplot2,scale_colour_gradientn)
importFrom(graphics,layout)
importFrom(igraph,"E<-")
importFrom(igraph,E)
importFrom(igraph,distances)
importFrom(igraph,is_tree)
importFrom(igraph,make_graph)
importFrom(lifecycle,deprecated)
importFrom(magrittr,"%>%")
importFrom(methods,as)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,slot)
importFrom(rSPDE,create_train_test_indices)
importFrom(rSPDE,gg_df)
importFrom(rSPDE,matern.covariance)
importFrom(rlang,.data)
importFrom(spatstat.geom,owin)
importFrom(spatstat.geom,psp)
importFrom(stats,approx)
importFrom(stats,as.formula)
importFrom(stats,delete.response)
importFrom(stats,deviance)
importFrom(stats,dist)
importFrom(stats,dnorm)
importFrom(stats,lm)
importFrom(stats,logLik)
importFrom(stats,model.matrix)
importFrom(stats,na.omit)
importFrom(stats,nobs)
importFrom(stats,optim)
importFrom(stats,pnorm)
importFrom(stats,predict)
importFrom(stats,profile)
importFrom(stats,qchisq)
importFrom(stats,rnorm)
importFrom(stats,rpois)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,simulate)
importFrom(stats,spline)
importFrom(stats,terms)
importFrom(stats,var)
importFrom(tidyr,drop_na)
importFrom(zoo,na.approx)
useDynLib(MetricGraph, .registration = TRUE)
