% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graph_lme.R
\name{glance.graph_lme}
\alias{glance.graph_lme}
\alias{glance}
\title{Glance at a \code{graph_lme} object}
\usage{
\method{glance}{graph_lme}(x, ...)
}
\arguments{
\item{x}{A \code{graph_lme} object.}

\item{...}{Additional arguments. Currently not used.}
}
\value{
A \code{\link[tidyr:tibble]{tidyr::tibble()}} with exactly one row and columns:
\itemize{
\item \code{nobs} Number of observations used.
\item \code{sigma} the square root of the estimated residual variance
\item \code{logLik} The log-likelihood of the model.
\item \code{AIC} Akaike's Information Criterion for the model.
\item \code{BIC} Bayesian Information Criterion for the model.
\item \code{deviance} Deviance of the model.
\item \code{df.residual} Residual degrees of freedom.
\item \code{model.type} Type of latent model fitted.
}
}
\description{
Glance accepts a \code{graph_lme} object and returns a
\code{\link[tidyr:tibble]{tidyr::tibble()}} with exactly one row of model summaries.
The summaries are the square root of the estimated variance of the measurement error, residual
degrees of freedom, AIC, BIC, log-likelihood,
the type of latent model used in the fit and the total number of observations.
}
\seealso{
\link{augment.graph_lme}
}
