% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{logo_lines}
\alias{logo_lines}
\title{Create lines for package name}
\usage{
logo_lines()
}
\value{
\code{SpatialLines} object with package name.
}
\description{
Create lines for package name
}
